@@ -1,13 +1,20 @@
 __all__ = [
     'ResultException', 'ResultType', 'Result', 'Ok', 'Err',
-    'WrappedBase', 'UnwrappedBase', 'wrap_result', 'auto_unwrap',
+    'WrappedBase', 'UnwrappedBase', 'wrap_result', 'auto_unwrap', 'unwrap',
 ]
 
 import types
 import asyncio
 import inspect
+import builtins
+import threading
 from typing import Any, Callable, ForwardRef
 from functools import wraps
+from contextlib import contextmanager
+
+
+# per-thread Result objects of each Result decorated function call
+_Thread_Err_map = {}
 
 
 class ResultException(Exception):
@@ -20,23 +27,56 @@ class _ResultType(type):
 
 class ResultType(_ResultType):
     def __getitem__(cls, key: Any) -> type:
+        V: type
+        E: type
+
         if cls is Ok or issubclass(cls, Ok):
+            V = key
+
+            if not isinstance(V, type):
+                raise TypeError(f'Type {V} sould be instance of type')
+
             T: ResultType = type(f'Ok[{key}]', (Ok,), {
                 '__annotations__': {
-                    'v': key
+                    'v': V
                 }
             })
         elif cls is Err or issubclass(cls, Err):
+            E = key
+
+            if not isinstance(E, type):
+                raise TypeError(f'Type {E} sould be instance of type')
+
             T: ResultType = type(f'Err[{key}]', (Err,), {
                 '__annotations__': {
-                    'e': key
+                    'e': E
                 }
             })
         elif cls is Result or issubclass(cls, Result):
+            if isinstance(key, (tuple, list)):
+                if len(key) == 1:
+                    V = key[0]
+                    E = Exception
+                elif len(key) == 2:
+                    V, E = key
+                else:
+                    raise TypeError(f'Type {key} sould be instance of type or tuple[type, type]')
+            elif isinstance(key, type):
+                V = key
+                E = Exception
+            else:
+                raise TypeError(f'Type {key} sould be instance of type or tuple[type, type]')
+
+            if not isinstance(V, type):
+                raise TypeError(f'Type {V} sould be instance of type')
+
+            if not isinstance(E, type):
+                raise TypeError(f'Type {E} sould be instance of type')
+
             T: ResultType = type(f'Result[{key}]', (Result,), {
                 '__annotations__': {
-                    'v': key[0],
-                    'e': key[1],
+                    'v': V,
+                    'e': E,
                 }
             })
         else:
@@ -49,6 +89,12 @@ class ResultType(_ResultType):
         V: type = cls.__annotations__['v']
         E: type = other.__annotations__['e']
 
+        if not isinstance(V, type):
+            raise TypeError(f'Type {V} sould be instance of type')
+
+        if not isinstance(E, type):
+            raise TypeError(f'Type {E} sould be instance of type')
+
         T: ResultType = type(f'Result[{V}, {E}]', (Result,), {
             '__annotations__': {
                 'v': V,
@@ -75,32 +121,73 @@ class ResultType(_ResultType):
             (fn := args[0]) and
             (callable(fn) or asyncio.iscoroutinefunction(fn))
         ):
-            V = cls.__annotations__['v']
-            E = cls.__annotations__['e']
+            V: type = cls.__annotations__['v']
+            E: type = cls.__annotations__['e']
+
+            if not isinstance(V, type):
+                raise TypeError(f'Type {V} sould be instance of type')
+
+            if not isinstance(E, type):
+                raise TypeError(f'Type {E} sould be instance of type')
             
             if asyncio.iscoroutinefunction(fn):
                 @wraps(fn)
                 async def wrap(*args, **kwargs) -> cls:
+                    global _Thread_Err_map
+                    r: Result
+
                     try:
                         v = await fn(*args, **kwargs)
-                        return Ok[V](v)
+                        r = Ok[V](v)
                     except Exception as e:
-                        return Err[E](e)
+                        r = Err[E](e)
+                    
+                    t = threading.current_thread()
+
+                    if t in _Thread_Err_map:
+                        _Thread_Err_map[t] = r
+                        r = r.unwrap()
+
+                    return r
             elif inspect.isfunction(fn) or inspect.ismethod(fn):
                 @wraps(fn)
                 def wrap(*args, **kwargs) -> cls:
+                    global _Thread_Err_map
+                    r: Result
+
                     try:
                         v = fn(*args, **kwargs)
-                        return Ok[V](v)
+                        r = Ok[V](v)
                     except Exception as e:
-                        return Err[E](e)
+                        r = Err[E](e)
+                    
+                    t = threading.current_thread()
+                    
+                    if t in _Thread_Err_map:
+                        _Thread_Err_map[t] = r
+                        r = r.unwrap()
+
+                    return r
             elif callable(fn):
                 def wrap(*args, **kwargs) -> cls:
+                    global _Thread_Err_map
+                    r: Result
+
                     try:
                         v = fn(*args, **kwargs)
-                        return Ok[V](v)
+                        r = Ok[V](v)
                     except Exception as e:
-                        return Err[E](e)
+                        r = Err[E](e)
+                    
+                    t = threading.current_thread()
+                    
+                    if t in _Thread_Err_map:
+                        _Thread_Err_map[t] = r
+                        r = r.unwrap()
+
+                    return r
+            else:
+                raise TypeError(f'Unsupported type of fn {fn}') # pragma: no cover
 
             return wrap
         else:
@@ -277,57 +364,98 @@ class UnwrappedBase:
         for a, v in state.items():
             setattr(self, a, v)
 
+#
+# unwrap
+#
+class Unwrap:
+    def __enter__(self):
+        global _Thread_Err_map
+        t = threading.current_thread()
+        _Thread_Err_map[t] = None
+        return self
+
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        global _Thread_Err_map
+        t = threading.current_thread()
+        del _Thread_Err_map[t]
+
+
+unwrap = Unwrap
+
+
+#
+# auto unwrap
+#
+_old_to_new_type_map = {}
+
+
+def _has_new_type(old_type):
+    global _old_to_new_type_map
+
+    if old_type in _old_to_new_type_map:
+        return True
+
+    return False
+
+
+def _get_new_type(old_type, default=NotImplemented):
+    global _old_to_new_type_map
+
+    if old_type in builtins.__dict__.values():
+        return old_type
+
+    if old_type in _old_to_new_type_map:
+        new_type = _old_to_new_type_map[old_type]
+        return new_type
+
+    if default is NotImplemented:
+        raise KeyError(f'missing key: {old_type}')
+
+    return default
+
+
+def _set_new_type(old_type, new_type):
+    global _old_to_new_type_map
+    _old_to_new_type_map[old_type] = new_type
+
 
 def _auto_unwrap_type(type_):
-    new_type_dict = {}
-    new_type = type(type_.__name__, (type_, UnwrappedBase), new_type_dict)
+    if _has_new_type(type_):
+        new_type = _get_new_type(type_)
+        return new_type
+    
+    new_type = types.new_class(type_.__name__, (type_, UnwrappedBase))
+    _set_new_type(type_, new_type)
 
     for a in dir(type_):
         f = getattr(type_, a)
 
         if not callable(f):
-            # print('not callable', a, f)
             continue
 
         if not hasattr(f, '__annotations__'):
-            # print('no __annotations__', a, f)
             continue
 
         return_type = f.__annotations__.get('return')
 
         if not return_type:
-            # print('no return_type', a, f)
-            continue  # pragma: no cover
-        
-        '''
-        if not (isinstance(return_type, type) or isinstance(return_type, str)):
-            print('return_type is not type/str', a, f, return_type, type(return_type))
             continue
 
-        if not (issubclass(return_type, Result) or isinstance(return_type, str)):
-            print('return_type is Result/str subclass', a, f, return_type, type(return_type))
-            continue
-        '''
-
-        '''
-        if isinstance(return_type, type) and issubclass(return_type, Result):
-            pass
-        elif isinstance(return_type, str):
-            pass
+        if _has_new_type(return_type):
+            new_return_type = _get_new_type(return_type)
         else:
-            print('return_type is not type Result/str', a, f, return_type, type(return_type))
-            continue
-        '''
+            new_return_type = auto_unwrap(return_type)
 
-        new_f = _auto_unwrap_func(f, old_type=return_type, new_type=new_type)
+        new_f = _auto_unwrap_func(f)
         setattr(new_type, a, new_f)
     
     return new_type
 
 
-def _auto_unwrap_func(func, old_type=None, new_type=None):
-    # print('auto_unwrap func-like', func, old_type, new_type)
-
+def _auto_unwrap_func(func):
+    global _old_to_new_type_map
+    
     @wraps(func)
     def wrapper(*args, **kwargs):
         if asyncio.iscoroutinefunction(func):
@@ -336,6 +464,9 @@ def _auto_unwrap_func(func, old_type=None, new_type=None):
 
                 if isinstance(r, Result):
                     r = r.unwrap()
+
+                old_type = type(r)
+                new_type = _get_new_type(old_type)
                 
                 if (inspect.isclass(old_type) and issubclass(old_type, WrappedBase) and isinstance(r, old_type)) or \
                    (isinstance(old_type, ForwardRef) and old_type.__forward_arg__ == r.__class__.__name__):
@@ -352,6 +483,9 @@ def _auto_unwrap_func(func, old_type=None, new_type=None):
 
             if isinstance(r, Result):
                 r = r.unwrap()
+
+            old_type = type(r)
+            new_type = _get_new_type(old_type)
             
             if (inspect.isclass(old_type) and issubclass(old_type, WrappedBase) and isinstance(r, old_type)) or \
                (isinstance(old_type, ForwardRef) and old_type.__forward_arg__ == r.__class__.__name__):
@@ -366,6 +500,9 @@ def _auto_unwrap_func(func, old_type=None, new_type=None):
 
             if isinstance(r, Result):
                 r = r.unwrap()
+
+            old_type = type(r)
+            new_type = _get_new_type(old_type)
             
             if (inspect.isclass(old_type) and issubclass(old_type, WrappedBase) and isinstance(r, old_type)) or \
                (isinstance(old_type, ForwardRef) and old_type.__forward_arg__ == r.__class__.__name__):
@@ -384,7 +521,10 @@ def auto_unwrap(obj) -> Any:
     This function auto unwraps Ok or Err type of Result
     '''
     if inspect.isclass(obj):
-        r = _auto_unwrap_type(obj)
+        if issubclass(obj, WrappedBase):
+            r = _auto_unwrap_type(obj)
+        else:
+            r = obj
     else:
         r = _auto_unwrap_func(obj)
     
